/* modified by azz: from... */

/*****************************************************************************/
/*									     */
/*									     */
/*	CP/M emulator version 0.1					     */
/*									     */
/*	written by Michael Bischoff (mbi@mo.math.nat.tu-bs.de)		     */
/*	June-1994							     */
/*									     */
/*	This file is distributed under the GNU COPYRIGHT		     */
/*	see COPYRIGHT.GNU for Copyright details				     */
/*									     */
/*									     */
/*****************************************************************************/

/* 8-Bit registers */
static const char *dtab1[] = { "B", "C", "D", "E", "H", "L", 0, "A" };
/* condition codes */
static const char *dtab4[] = { "NZ", "Z", "NC", "C", "PO", "PE", "P", "M" };
/* opcodes */
static const char *dtab5[] = { "ADD A,", "ADC A,", "SUB ", "SBC A,", "AND ", "XOR ", "OR ", "CP " };
static const char *dtabix[] = { "POP %s", "EX (SP),%s", "PUSH %s", "JP (%s)",
				    "EX DE,%s", "LD SP,%s" };
static const char *dtab2[] = { "BC", "DE", 0, "SP" };
/*static const char *dtab3[] = { "BC", "DE", 0, "AF" }; */

static const char *op1tab[] = {	/* 0s are handled explicitly */
    "NOP",       0,        "LD (BC),A", 0, 0, 0, 0, "RLCA",
    "EX AF,AF", 0,        "LD A,(BC)", 0, 0, 0, 0, "RRCA",
    "DJNZ %04xh", 0,        "LD (DE),A", 0, 0, 0, 0, "RLA",
    "JR %04xh",   0,        "LD A,(DE)", 0, 0, 0, 0, "RRA",
    "JR NZ,%04xh",0,	      0,        0, 0, 0, 0, "DAA",
    "JR Z,%04xh", 0,        0, 	   0, 0, 0, 0, "CPL",
    "JR NC,%04xh",0,        0,	   0, 0, 0, 0, "SCF",
    "JR C,%04xh", 0,	      0,	   0, 0, 0, 0, "CCF" };

static const char *op2tab[] = {
    0, "POP BC", 0, "JP %04xh",      0, "PUSH BC", 0, 0,   0, "RET",     0, "???",         0, "CALL %04xh", 0, 0,
    0, "POP DE", 0, "OUT (%03xh),A", 0, "PUSH DE", 0, 0,   0, "EXX",     0, "IN A,(%03xh)", 0, "???",       0, 0,
    0, "POP HL", 0, "EX (SP),HL",   0, "PUSH HL", 0, 0,   0, "JP (HL)", 0, "EX DE,HL",    0, "???",       0, 0,
    0, "POP AF", 0, "DI",           0, "PUSH AF", 0, 0,   0, "LD SP,HL",0, "EI",          0, "???",       0, 0 };
static const char *cbtab1[] = {
    "RLC", "RRC", "RL", "RR", "SLA", "SRA", "? SLIA", "SRL" };
static const char *cbtab2[] = {
    0, "BIT", "RES", "SET" };
static unsigned char ixy_possible8[32] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 0x00,
    0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0xbf, 0x40,
    0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static char edcmds[] = {
    0x44, 0xa0, 0xa8, 0xb0, 0xb8, 0x47, 0x4f, 0x57,
    0x5f, 0xa1, 0xb1, 0xa9, 0xb9, 0x46, 0x56, 0x5e,
    0x67, 0x6f, 0xa3, 0xab, 0xb3, 0xbb, 0x45, 0x4d,
0 };
static const char *edtxt[] = {
    "NEG", "LDI", "LDD", "LDIR", "LDDR", "LD I,A", "LD R,A", "LD A,I",
    "LD A,R", "CPI", "CPIR", "CPD", "CPDR", "IM 0", "IM 1", "IM 2", "RRD",
    "RLD", "OUTI", "OUTD", "OTIR", "OTDR", "RETN", "RETI"
    };

static char incregcmds[] = { 0x40, 0x48, 0x50, 0x58, 0x60, 0x68, 0 };
static const char *incregtxt[] = { "B", "C", "D", "E", "H", "L" };

static const unsigned char *edcommand(const unsigned char *PC, char *s) {
    unsigned o, w;
    const char *dr;
    o = *PC;
    
    dtab2[2] = "HL";
    dr = dtab2[(o & 0x30) >> 4];

    if (strchr(incregcmds,o)) {
      sprintf(s, "IN %s,(C)", incregtxt[strchr(incregcmds,o)-incregcmds]);
      return PC+1;
    };

    if (strchr(incregcmds,o-1)) {
      sprintf(s, "OUT (C),%s", incregtxt[strchr(incregcmds,o-1)-incregcmds]);
      return PC+1;
    };

    switch (o & 0xcf) {
    case 0x42:
	sprintf(s, "SBC HL,%s", dr);
	break;
    case 0x43:
	w = *++PC;
	w += *++PC << 8;
	sprintf(s, "LD (%04xh),%s", w, dr);
	break;
    case 0x4a:
	sprintf(s, "ADC HL,%s", dr);
	break;
    case 0x4b:
	w = *++PC;
	w += *++PC << 8;
	sprintf(s, "LD %s,(%04xh)", dr, w);
	break;
    default:
	if (o && (dr = strchr(edcmds, o))) {
	    strcpy(s, edtxt[dr-edcmds]);
	    break;
	}
	sprintf(s, "DB 0edh");
	return PC;
    }
    return PC+1;
}


/* PC is only needed for JR instructions */
const unsigned char *disz80(const unsigned char *PC, char *s, unsigned pc) {
    unsigned o, m, r;
    const char *regh, *regl;
    char ir[3];
    char i_offset[8];
    unsigned w;
    
    dtab1[6] = "(HL)";
    strcpy(ir,"HL");
    o = *PC;
 redo:
    if (ir[0]==0) ir[0]='I'; /* I have no idea why this is necessary. -azz */
    regh = dtab1[m = (o >> 3) & 7];
    regl = dtab1[r = o & 7];
    
    switch (o >> 6) {
    case 0:		/* the lower block */
	switch(o & 7) {
	case 4:
	    sprintf(s, "INC %s", regh);
	    break;
	case 5:
	    sprintf(s, "DEC %s", regh);
	    break;
	case 6:
	    sprintf(s, "LD %s,0%02xh", regh, *++PC);
	    break;
	default:
	    dtab2[2] = ir;
	    switch (o) {
	    case 0x01: case 0x11: case 0x21: case 0x31:
		w = *++PC;
		w += *++PC << 8;
		sprintf(s, "LD %s,%04xh", dtab2[o>>4], w);
		break;
	    case 0x03: case 0x13: case 0x23: case 0x33:
		sprintf(s, "INC %s", dtab2[o>>4]);
		break;
	    case 0x0b: case 0x1b: case 0x2b: case 0x3b:
		sprintf(s, "DEC %s", dtab2[o>>4]);
		break;
	    case 0x09: case 0x19: case 0x29: case 0x39:
		sprintf(s, "ADD %s,%s", ir, dtab2[o>>4]);
		break;
	    case 0x22: case 0x2a: case 0x32: case 0x3a:
		/* commands which use a 16 bit argument */
		w = *++PC;
		w += *++PC << 8;
		switch (o) {
		case 0x22:
		    sprintf(s, "LD (%04xh),%s", w, ir);
		    break;
		case 0x2a:
		    sprintf(s, "LD %s,(%04xh)", ir, w);
		    break;
		case 0x32:
		    sprintf(s, "LD (%04xh),A", w);
		    break;
		case 0x3a:
		    sprintf(s, "LD A,(%04xh)", w);
		    break;
		}
		break;
	    case 0x10: case 0x18: case 0x20:
	    case 0x28: case 0x30: case 0x38:
		/* jump relative */
#if 1
		if ((w = *++PC) & 0x80)
		    w -= 256;
		w += pc + 2;
#else
		w = *++PC;	/* no service for destination address */
#endif
		sprintf(s, op1tab[o], w);
		break;
	    default:
		strcpy(s, op1tab[o]);
		break;
	    }
	}
	break;	/* lower block done */
    case 1:		/* second block is easy */
	if (o != 0x76)
	    sprintf(s, "LD %s,%s", regh, regl);
	else
	    strcpy(s, "HALT");
	break;
    case 2:		/* third block is easy */
	strcpy(s, dtab5[m]);
	strcat(s, regl);
	break;
    case 3:		/* fourth block */
	switch(o & 7) {
	case 0:
	    sprintf(s, "RET %s", dtab4[m]);
	    break;
	case 2:
	    w = *++PC;
	    w += *++PC << 8;
	    sprintf(s, "JP %s,%04xh", dtab4[m], w);
	    break;
	case 4:
	    w = *++PC;
	    w += *++PC << 8;
	    sprintf(s, "CALL %s,%04xh", dtab4[m], w);
	    break;
	case 6:
	    sprintf(s, "%s0%02xh", dtab5[m], *++PC);
	    break;
	case 7:
	    sprintf(s, "RST %03xH", m << 3);
	    break;
	default:
	    switch (o) {
	    case 0xc3: case 0xcd:
		w = *++PC;
		w += *++PC << 8;
		break;
	    case 0xd3: case 0xdb:
		w = *++PC;
		break;

	    /* SPECIAL Z80 COMMANDS START HERE: */
	    case 0xcb:	/* rotate-commands */
	    ddcb:
		o = *++PC;
		regl = dtab1[r = o & 7];
		m = (o >> 3) & 7;
		if (o & 0xc0)
		    sprintf(s, "%s %d,%s", cbtab2[o >> 6], m, dtab1[r]);
		else
		    sprintf(s, "%s %s", cbtab1[m], dtab1[r]);
		return PC+1;
	    case 0xed:
		return edcommand(PC+1, s);
	    case 0xdd:
		strcpy(ir,"IX");
		goto ixiy;
	    case 0xfd:
		strcpy(ir,"IY");
	    ixiy:
		dtab1[6] = i_offset;
		m = PC[2];	/* offset */
		sprintf(i_offset, "(%s%c%02xh)", ir, m&0x80 ? '-' : '+',
			m&0x80 ? 256-m : m);
		o = *++PC;
		if (o == 0xcb) {
		    /* rotate with index reg */
		    o = PC[2];
		    if ((o & 7) != 6) { /* bad opcode */
		      sprintf(s,"DB %03xh, %03xh, %03xh",PC[0],PC[1],PC[2]);
		      return PC+3;
		    }
		    /* return PC-1; */	/* invalid opcode! */
		    ++PC;	/* skip offset */
		    goto ddcb;
		}
		if (ixy_possible8[o >> 3] & (1<<(o&7))) {
		    /* is a command which addresses 8 bit index+offset */
		    ++PC;	/* skip offset */
		    goto redo;
		}
		{   unsigned char *p, xtra1[] = {
		    0x21, 0x22, 0x2a, 0x09, 0x19, 0x29, 0x39, 0x23, 0x2b, 0
		    }, xtra2[] = {
		    0xe1, 0xe3, 0xe5, 0xe9, 0xeb, 0xf9, 0 };
		    if (o) {
			if ((p = strchr(xtra1, o)))
			    goto redo;
			else if ((p = strchr(xtra2, o))) {
			    if (ir[0]=='\0') ir[0]='I';
			    sprintf(s, dtabix[p-xtra2], ir);
			    return PC+1;
			}
		    }
		}
		sprintf(s, "DB %03xh", o);
		return PC+1;
		/* return PC-1;	*/ /* unknown opcode */
	    }
	    sprintf(s, op2tab[o & 63], w);
	    break;
	}
	break;
    }
    return PC+1;
}


